import functools, inspect, io, itertools, os, pprint, time, unittest

from . import module
from .multimedia import ffmpeg
from .string import snake_to_camel
from .types import Missing

_all = module.All(globals())


def with_tags(tags=None, required=None):
    def _with_tags(f):
        f.tags = set([tags] if isinstance(tags, str) else tags or [])
        f.required_tags = set([required] if isinstance(required, str) else required or [])
        return f
    return _with_tags


class InMixin(object):

    @staticmethod
    def assert_in_hook(obj):
        return obj if isinstance(obj, (str, bytes)) else sorted(obj)

    def assertIn(self, obj_a, obj_b, msg=None):  # pylint:disable=invalid-name
        assert obj_a in obj_b, f"{obj_a} not in {self.assert_in_hook(obj_b)}: {msg or ''}"

    def assertNotIn(self, obj_a, obj_b, msg=None):  # pylint:disable=invalid-name
        assert obj_a not in obj_b, f"{obj_a} in {self.assert_in_hook(obj_b)}: {msg or ''}"


class InspectMixin(object):

    @property
    def current_test(self):
        return getattr(self, self.id().split('.')[-1])

    @classmethod
    def get_test_methods(cls):
        return (
            (n, m) for n, m in inspect.getmembers(cls)
            if n.startswith('test_') and hasattr(m, '__call__')
        )


class AwareTearDownMixin(object):

    def awareTearDown(self, result):  # pylint:disable=invalid-name
        pass  # de bleu, c'est fantastique !

    def run(self, result=None):
        result = super().run(result)
        self.awareTearDown(result)
        return result


class FilterByTagsMixin(InspectMixin):
    """
    Allow to filter unit-tests by tags, including by default, the `TestCaseName` and
    `TestCaseName.method_name`.
    """

    tags = required_tags = ()
    fast_class_skip_enabled = False
    extra_tags_variable = 'TEST_EXTRA_TAGS'
    only_tags_variable = 'TEST_ONLY_TAGS'
    skip_tags_variable = 'TEST_SKIP_TAGS'

    @staticmethod
    def should_run(tags, required_tags, extra_tags, only_tags, skip_tags):
        all_tags = tags | required_tags
        if all_tags & skip_tags:
            return False
        if only_tags and not all_tags & only_tags:
            return False
        if required_tags and not required_tags & (only_tags | extra_tags):
            return False
        return True

    @classmethod
    def setUpClass(cls):  # pylint:disable=invalid-name
        if cls.fast_class_skip_enabled:
            cls.fast_class_skip()
        super().setUpClass()

    @classmethod
    def fast_class_skip(cls):
        methods = cls.get_test_methods()
        should_run = functools.partial(
            cls.should_run,
            extra_tags=cls.get_extra_tags(),
            only_tags=cls.get_only_tags(),
            skip_tags=cls.get_skip_tags()
        )
        if not any(should_run(cls.get_tags(m), cls.get_required_tags(m)) for n, m in methods):
            raise unittest.SkipTest('Test skipped by FilterByTagsMixin.fast_class_skip')

    @classmethod
    def get_extra_tags(cls):
        return set(t for t in os.environ.get(cls.extra_tags_variable, '').split(',') if t)

    @classmethod
    def get_only_tags(cls):
        return set(t for t in os.environ.get(cls.only_tags_variable, '').split(',') if t)

    @classmethod
    def get_skip_tags(cls):
        return set(t for t in os.environ.get(cls.skip_tags_variable, '').split(',') if t)

    @classmethod
    def get_tags(cls, current_test):
        my_id = (cls.__name__, current_test.__name__)
        return set(itertools.chain(
            cls.tags,
            getattr(current_test, 'tags', ()),
            (my_id[0], '.'.join(my_id)))
        )

    @classmethod
    def get_required_tags(cls, current_test):
        return set(itertools.chain(cls.required_tags, getattr(current_test, 'required_tags', ())))

    def setUp(self):  # pylint:disable=invalid-name
        if not self.should_run(
            self.get_tags(self.current_test),
            self.get_required_tags(self.current_test),
            self.get_extra_tags(),
            self.get_only_tags(),
            self.get_skip_tags()
        ):
            raise unittest.SkipTest('Test skipped by FilterByTagsMixin.setUp')
        super().setUp()


class FFmpegMixin(object):

    ffmpeg_class = ffmpeg.FFmpeg

    @classmethod
    def setUpClass(cls):  # pylint:disable=invalid-name
        for name, stream_class in cls.ffmpeg_class.ffprobe_class.stream_classes.items():
            assert stream_class is not None, name
            assert stream_class.codec_class is not None, name
        super().setUpClass()

    def setUp(self):  # pylint:disable=invalid-name
        super().setUp()
        self.ffmpeg = self.ffmpeg_class()
        self.ffprobe = self.ffmpeg.ffprobe_class()

    # Codecs Asserts

    def assertMediaCodecEqual(
        self,
        path,
        stream_type,
        index,
        **codec_attrs
    ):  # pylint:disable=invalid-name
        codec = getattr(self.ffprobe, f'get_{stream_type}_streams')(path)[index].codec
        for attr, value in codec_attrs.items():
            self.assertEqual(getattr(codec, attr), value, msg=f'Codec attribute {attr}')

    def assertAudioCodecEqual(self, path, index, **codec_attrs):  # pylint:disable=invalid-name
        self.assertMediaCodecEqual(path, 'audio', index, **codec_attrs)

    def assertSubtitleCodecEqual(self, path, index, **codec_attrs):  # pylint:disable=invalid-name
        self.assertMediaCodecEqual(path, 'subtitle', index, **codec_attrs)

    def assertVideoCodecEqual(self, path, index, **codec_attrs):  # pylint:disable=invalid-name
        self.assertMediaCodecEqual(path, 'video', index, **codec_attrs)

    # Streams Asserts

    def assertAudioStreamEqual(  # pylint:disable=invalid-name
        self,
        first_path,
        second_path,
        first_index,
        second_index,
        same_codec=True
    ):
        first = self.ffprobe.get_audio_streams(first_path)[first_index]
        second = self.ffprobe.get_audio_streams(second_path)[second_index]
        if same_codec:
            self.assertEqual(first.codec, second.codec, msg='Codec mistmatch.')
        self.assertEqual(first.bit_rate, second.bit_rate, msg='Bit rate mistmatch.')

    def assertMediaFormatEqual(  # pylint:disable=invalid-name
        self,
        first_path,
        second_path,
        same_bit_rate=True,
        same_duration=True,
        same_size=True,
        same_start_time=True
    ):
        formats = [self.ffprobe.get_media_info(p)['format'] for p in (first_path, second_path)]
        bit_rates, durations, sizes, start_times = [], [], [], []
        for the_format in formats:
            the_format.pop('filename')
            the_format.pop('tags', None)
            durations.append(float(the_format.pop('duration')))
            bit_rates.append(float(the_format.pop('bit_rate', 0)))
            sizes.append(int(the_format.pop('size')))
            start_times.append(float(the_format.pop('start_time')))
        if same_bit_rate:
            self.assertRelativeEqual(*bit_rates, msg='Bit rate mistmatch.')
        if same_duration:
            self.assertRelativeEqual(*durations, msg='Duration mistmatch.')
        if same_size:
            self.assertRelativeEqual(*sizes, msg='Size mistmatch.')
        if same_start_time:
            self.assertRelativeEqual(*start_times, msg='Start time mistmatch.')
        self.assertDictEqual(*formats)

    def assertVideoStreamEqual(  # pylint:disable=invalid-name
        self,
        first_path,
        second_path,
        first_index,
        second_index,
        same_codec=True
    ):
        first = self.ffprobe.get_video_streams(first_path)[first_index]
        second = self.ffprobe.get_video_streams(second_path)[second_index]
        if same_codec:
            self.assertEqual(first.codec, second.codec, msg='Codec mismatch.')
        self.assertRelativeEqual(
            first.avg_frame_rate, second.avg_frame_rate, msg='Average frame rate mismatch.')
        if first.nb_frames:
            self.assertRelativeEqual(
                first.nb_frames, second.nb_frames, msg='Number of frames mistmatch.')
        self.assertEqual(first.height, second.height, msg='Height mismatch.')
        self.assertEqual(first.width, second.width, msg='Width mismatch.')

    # Encoding Asserts

    def assertEncodeFailure(self, generator):  # pylint:disable=invalid-name
        return self.assertEncodeState(generator, state=ffmpeg.EncodeState.FAILURE)

    def assertEncodeSuccess(self, generator):  # pylint:disable=invalid-name
        return self.assertEncodeState(generator, state=ffmpeg.EncodeState.SUCCESS)

    def assertEncodeState(self, generator, state):  # pylint:disable=invalid-name
        results = list(generator)
        result = io.StringIO()
        statistics = results[-1]
        pprint.pprint(
            {a: getattr(statistics, a) for a in dir(statistics) if a[0] != '_'},
            stream=result)
        self.assertEqual(statistics.state, state, result.getvalue())
        return results


class MissingMixin(object):

    def assertIsMissing(self, value, *args, **kwargs):  # pylint:disable=invalid-name
        return self.assertIs(value, Missing, *args, **kwargs)

    def assertIsNotMissing(self, value, *args, **kwargs):  # pylint:disable=invalid-name
        return self.assertIsNot(value, Missing, *args, **kwargs)


class SnakeCaseMixin(object):

    def __getattr__(self, name):
        if name.lower() == name:
            attribute = snake_to_camel(name if name.startswith('assert_') else f'assert_{name}')
            return getattr(self, attribute)
        raise AttributeError


class TimingMixin(object):

    timing_logger = None

    def setUp(self):  # pylint:disable=invalid-name
        self.start_time = time.time()
        super().setUp()

    def tearDown(self):  # pylint:disable=invalid-name
        super().tearDown()
        if self.timing_logger:
            self.timing_logger.info(f'{self.id()}: {time.time() - self.start_time:.3f}')


class Asserts(InMixin, MissingMixin, SnakeCaseMixin, unittest.TestCase):

    def runTest(self, *args, **kwargs):  # pylint:disable=invalid-name
        pass


asserts = Asserts()

__all__ = _all.diff(globals())



import numbers, re

from pytoolbox import module

_all = module.All(globals())

BIT_RATE_REGEX = re.compile(r'^(?P<value>\d+\.?\d*)(?P<units>[a-z]+)/s$')
BIT_RATE_COEFFICIENT_FOR_UNIT = {'b': 1, 'k': 1000, 'm': 1000**2, 'g': 1000**3}
PIPE_REGEX = re.compile(r'^-$|^pipe:\d+$')
SIZE_REGEX = re.compile(r'^(?P<value>\d+\.?\d*)(?P<units>[a-zA-Z]+)$')
SIZE_COEFFICIENT_FOR_UNIT = {'b': 1, 'k': 1024, 'm': 1024**2, 'g': 1024**3}
WIDTH, HEIGHT = range(2)  # noqa


def is_pipe(path):
    return isinstance(path, str) and bool(PIPE_REGEX.match(path))


def to_bit_rate(bit_rate):
    match = BIT_RATE_REGEX.match(bit_rate)
    if match:
        match = match.groupdict()
        return int(float(match['value']) * BIT_RATE_COEFFICIENT_FOR_UNIT[match['units'][0]])
    if bit_rate == 'N/A':
        return None
    raise ValueError(bit_rate)


def to_frame_rate(frame_rate):
    if isinstance(frame_rate, numbers.Number):
        return frame_rate
    if '/' in frame_rate:
        try:
            num, denom = frame_rate.split('/')
            return float(num) / float(denom)
        except ZeroDivisionError:
            return None
    return float(frame_rate)


def to_size(size):
    match = SIZE_REGEX.match(size)
    if match:
        match = match.groupdict()
        return int(float(match['value']) * SIZE_COEFFICIENT_FOR_UNIT[match['units'][0].lower()])
    raise ValueError(size)


__all__ = _all.diff(globals())



__all__ = ['log_to_console']


def log_to_console(settings):
    """
    Update settings to make all loggers use the console.

    **Example usage**

    >>> import collections
    >>> settings = collections.namedtuple('settings', ['DEBUG', 'LOGGING'])
    >>> settings.DEBUG = True
    >>> settings.LOGGING = {
    ...     'version': 1,
    ...     'loggers': {
    ...         'django': {
    ...             'handlers': ['file'], 'level': 'INFO', 'propagate': True
    ...         },
    ...         'django.request': {
    ...             'handlers': ['mail_admins'], 'level': 'ERROR', 'propagate': True
    ...         },
    ...         'mysite': {
    ...             'handlers': ['console'], 'level': 'INFO', 'propagate': True
    ...         }
    ...     }
    ... }
    >>> expected_settings = collections.namedtuple('settings', ['DEBUG', 'LOGGING'])
    >>> expected_settings.DEBUG = True
    >>> expected_settings.LOGGING = {
    ...     'version': 1,
    ...     'loggers': {
    ...         'django': {
    ...             'handlers': ['console'], 'level': 'INFO', 'propagate': True
    ...         },
    ...         'django.request': {
    ...             'handlers': ['console'], 'level': 'ERROR', 'propagate': True
    ...         },
    ...         'mysite': {
    ...             'handlers': ['console'], 'level': 'INFO', 'propagate': True
    ...         }
    ...     }
    ... }
    >>> log_to_console(settings)
    >>> settings.LOGGING == expected_settings.LOGGING
    True
    """
    for logger in settings.LOGGING['loggers']:
        settings.LOGGING['loggers'][logger]['handlers'] = ['console']



from ipaddress import ip_address

from pytoolbox import exceptions

__all__ = ['ip_address', 'IPSocket']


def IPSocket(socket):  # pylint:disable=invalid-name
    """
    This helper create a dictionary containing address and port from a parsed IP address string.
    Throws InvalidIPSocketError in case of failure.

    **Example usage**

    >>> IPSocket('gaga:gogo')
    Traceback (most recent call last):
        ...
    pytoolbox.exceptions.InvalidIPSocketError: gaga:gogo is not a valid IP socket.
    >>>
    >>> from pytoolbox.unittest import asserts
    >>> asserts.dict_equal(
    ...     IPSocket('239.232.0.222:5004'),
    ...     {'ip': '239.232.0.222', 'port': 5004})

    .. warning::

        TODO IPv6 ready : >>> IPSocket('[2001:0db8:0000:0000:0000:ff00:0042]:8329')
    """
    try:
        (address, port) = socket.rsplit(':', 1)
        # address = address.translate(None, '[]')
        ip_address(address)  # Seem not IPv6 ready
        port = int(port)
    except Exception as e:
        raise exceptions.InvalidIPSocketError(socket=socket) from e
    return {'ip': address, 'port': port}



from pytoolbox.selenium import Keys

__all__ = ['BootstrapSwitchMixin']


class BootstrapSwitchMixin(object):

    component = 'bootstrapSwitch'
    key_map = {True: Keys.RIGHT, False: Keys.LEFT}

    @staticmethod
    def clean_value(value):
        return {'on': True, 'off': False, True: True, False: False}[value]

    def send_keys(self, *value):
        return super().send_keys(*[self.key_map.get(v, v) for v in value])



import datetime, errno, itertools, json, math, os, re, subprocess
from pathlib import Path

from xml.dom import minidom

from pytoolbox import subprocess as py_subprocess
from pytoolbox.datetime import parts_to_time, secs_to_time
from . import miscellaneous, utils

__all__ = ['DURATION_REGEX', 'FFprobe']

DURATION_REGEX = re.compile(r'PT(?P<hours>\d+)H(?P<minutes>\d+)M(?P<seconds>[^S]+)S')


class FFprobe(object):

    executable = 'ffprobe'
    duration_regex = DURATION_REGEX
    format_class = None
    media_class = miscellaneous.Media
    stream_classes = {'audio': None, 'subtitle': None, 'video': None}

    def __init__(self, executable=None):
        self.executable = executable or self.executable

    def __call__(self, *arguments):
        """Call FFprobe with given arguments and return the output (unicode string)."""
        process = py_subprocess.raw_cmd(
            itertools.chain([self.executable], arguments),
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            universal_newlines=True)
        process.wait()
        return process.stdout.read()

    def get_media_duration(self, media, as_delta=False, fail=False):
        """
        Returns the duration of a media as an instance of time or None in case of error.

        Set `media` to an instance of `self.media_class`, a path or the output of
        `get_media_info()`.

        If `media` is the path to a MPEG-DASH MPD, then duration will be parser from value of key
        *mediaPresentationDuration*.
        """
        if isinstance(media, (str, Path)) and os.path.splitext(media)[1] == '.mpd':
            with open(media, encoding='utf-8') as f:
                mpd = minidom.parse(f)
            if mpd.firstChild.nodeName == 'MPD':
                match = self.duration_regex.search(
                    mpd.firstChild.getAttribute('mediaPresentationDuration'))
                if match is not None:
                    hours, minutes = int(match.group('hours')), int(match.group('minutes'))
                    microseconds, seconds = math.modf(float(match.group('seconds')))
                    microseconds, seconds = int(1000000 * microseconds), int(seconds)
                    return parts_to_time(hours, minutes, seconds, microseconds, as_delta=as_delta)
        else:
            info = self.get_media_info(media, fail)
            duration = None
            if info:
                try:
                    duration = secs_to_time(float(info['format']['duration']), as_delta=as_delta)
                except KeyError:
                    return None
            # ffmpeg may return this so strange value, 00:00:00.04, let it being None
            if duration and (duration >= datetime.timedelta(seconds=1) if as_delta else
                             duration >= datetime.time(0, 0, 1)):
                return duration
        return None

    def get_media_info(self, media, fail=False):
        """
        Return a Python dictionary containing information about the media or None in case of error.
        Set `media` to an instance of `self.media_class` or a path.
        If `media` is a Python dictionary, then it is returned.
        """
        if isinstance(media, dict):
            return media
        media = self.to_media(media)
        try:
            if utils.is_pipe(media.path):
                raise NotImplementedError('Read media information from a PIPE not yet implemented.')

            return json.loads(
                subprocess.check_output([
                    self.executable,
                    '-v', 'quiet',
                    '-print_format', 'json',
                    '-show_format',
                    '-show_streams',
                    media.path
                ]).decode('utf-8'))
        except OSError as e:
            # Executable does not exist
            if fail or e.errno == errno.ENOENT:
                raise
        except Exception:  # pylint:disable=broad-except
            if fail:
                raise
        return None

    def get_media_format(self, media, fail=False):
        """
        Return information about the container (and file) or None in case of error.
        Set `media` to an instance of `self.media_class`, a path or the output of
        `get_media_info()`.
        """
        info = self.get_media_info(media, fail)
        try:
            cls, the_format = self.format_class, info['format']
            if cls and not isinstance(the_format, cls):  # pylint:disable=all
                return cls(the_format)                   # pylint:disable=not-callable
            return the_format
        except Exception:  # pylint:disable=broad-except
            if fail:
                raise
        return None

    def get_media_streams(self, media, condition=lambda stream: True, fail=False):
        """
        Return a list with the media streams of `media` or [] in case of error.
        Set `media` to an instance of `self.media_class`, a path or the output of
        `get_media_info()`.
        """
        info = self.get_media_info(media, fail)
        try:
            raw_streams = (s for s in info['streams'] if condition(s))
        except Exception:  # pylint:disable=broad-except
            if fail:
                raise
            return []
        streams = []
        for stream in raw_streams:
            stream_class = self.stream_classes[stream['codec_type']]
            streams.append(
                stream_class(stream) if stream_class and not isinstance(stream, stream_class)
                else stream)
        return streams

    def get_audio_streams(self, media, fail=False):
        """
        Return a list with the audio streams of `media` or [] in case of error.
        Set `media` to an instance of `self.media_class`, a path or the output of
        `get_media_info()`.
        """
        return self.get_media_streams(
            media,
            condition=lambda s: s['codec_type'] == 'audio',
            fail=fail)

    def get_subtitle_streams(self, media, fail=False):
        """
        Return a list with the subtitle streams of `media` or [] in case of error.
        Set `media` to an instance of `self.media_class`, a path or the output of
        `get_media_info()`.
        """
        return self.get_media_streams(
            media,
            condition=lambda s: s['codec_type'] == 'subtitle',
            fail=fail)

    def get_video_streams(self, media, fail=False):
        """
        Return a list with the video streams of `media` or [] in case of error.
        Set `media` to an instance of `self.media_class`, a path or the output of
        `get_media_info()`.
        """
        return self.get_media_streams(
            media,
            condition=lambda s: s['codec_type'] == 'video',
            fail=fail)

    def get_video_frame_rate(self, media, index=0, fail=False):
        """
        Return the frame rate of the video stream at `index` in `media` or None in case of error.
        Set `media` to an instance of `self.media_class`, a path or the output of
        `get_media_info()`.
        """
        try:
            stream = self.get_video_streams(media)[index]
            if isinstance(stream, dict):
                return utils.to_frame_rate(stream['avg_frame_rate'])
            else:
                return stream.avg_frame_rate
        except Exception:  # pylint:disable=broad-except
            if fail:
                raise
        return None

    def get_video_resolution(self, media, index=0, fail=False):
        """
        Return [width, height] of the video stream at `index` in `media` or None in case of error.
        Set `media` to an instance of `self.media_class`, a path or the output of
        `get_media_info()`.
        """
        try:
            stream = self.get_video_streams(media)[index]
            is_dict = isinstance(stream, dict)
            if is_dict:
                return [int(stream['width']), int(stream['height'])]
            return [stream.width, stream.height]
        except Exception:  # pylint:disable=broad-except
            if fail:
                raise
        return None

    def to_media(self, media):
        return media if isinstance(media, self.media_class) else self.media_class(media)



from pytoolbox.enum import OrderedEnum
from . import tag

__all__ = ['ExposureMode', 'Photo']


class ExposureMode(OrderedEnum):
    AUTO = 0
    MANUAL = 1
    BRACKET = 2


class Photo(tag.TagSet):

    @property
    def date(self):
        return self.metadata.get_date()

    @property
    def exposure_mode(self):
        return ExposureMode(self.metadata['Exif.Photo.ExposureMode'].data)

    @property
    def exposure_time(self):
        return self.metadata['Exif.Photo.ExposureTime'].data

    @property
    def fnumber(self):
        return self.clean_number(self.metadata['Exif.Photo.FNumber'].data)

    @property
    def focal_length(self):
        return self.clean_number(self.metadata['Exif.Photo.FocalLength'].data)

    @property
    def iso_speed(self):
        return self.clean_number(self.metadata['Exif.Photo.ISOSpeedRatings'].data)

    @property
    def sensing_method(self):
        return self.metadata['Exif.Photo.SensingMethod'].data

    @property
    def white_balance(self):
        return self.metadata['Exif.Photo.WhiteBalance'].data



import math, re

from . import module

_all = module.All(globals())

DEFAULT_BITRATE_UNITS = ('bit/s', 'kb/s', 'Mb/s', 'Gb/s', 'Tb/s', 'Pb/s', 'Eb/s', 'Zb/s', 'Yb/s')
DEFAULT_FILESIZE_ARGS = {
    'gnu': {'base': 1000, 'units': ('B', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')},
    'nist': {'base': 1024, 'units': ('B', 'kB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB')},
    'si': {'base': 1000, 'units': ('B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB')},
}
DEFAULT_FREQUENCY_UNITS = ('Hz', 'kHz', 'MHz', 'GHz', 'THz', 'PHz', 'EHz', 'ZHz', 'YHz')
DIGIT_REGEX = re.compile(r'(\d+)')


def _naturalnumber(
    number,
    base,
    units,
    format='{sign}{value:.3g} {unit}',
    scale=None
):  # pylint:disable=redefined-builtin
    sign, number = '' if number >= 0 else '-', abs(number)
    if scale is None:
        scale = min(int(math.log(max(1, number), base)), len(units) - 1)
    unit = units[scale]
    value = number / (base ** scale)
    return format.format(sign=sign, value=value, unit=unit)


def naturalbitrate(
    bps,
    format='{sign}{value:.3g} {unit}',
    scale=None,
    units=DEFAULT_BITRATE_UNITS
):  # pylint:disable=redefined-builtin
    """
    Return a human readable representation of a bit rate taking `bps` as the rate in bits/s.

    The unit is taken from:

    * The `scale` if not None (0=bit/s, 1=kb/s, 2=Mb/s, ...).
    * The right scale from `units`.

    **Example usage**

    >>> naturalbitrate(-10)
    '-10 bit/s'
    >>> naturalbitrate(0.233)
    '0.233 bit/s'
    >>> naturalbitrate(69.5, format='{value:.2g} {unit}')
    '70 bit/s'
    >>> naturalbitrate(999.9, format='{value:.0f}{unit}')
    '1000bit/s'
    >>> naturalbitrate(1060)
    '1.06 kb/s'
    >>> naturalbitrate(3210837)
    '3.21 Mb/s'
    >>> naturalbitrate(16262710, units=['bps', 'Kbps'])
    '1.63e+04 Kbps'
    >>> naturalbitrate(3210837, scale=1, format='{value:.2f} {unit}')
    '3210.84 kb/s'
    """
    return _naturalnumber(bps, base=1000, format=format, scale=scale, units=units)


def naturalfilesize(
    bytes,
    system='nist',
    format='{sign}{value:.3g} {unit}',
    scale=None,
    args=DEFAULT_FILESIZE_ARGS
):  # pylint:disable=dangerous-default-value,redefined-builtin
    """
    Return a human readable representation of a *file* size taking `bytes` as the size in bytes.

    The base and units taken from:

    * The value in `args` with key `system` if not None.
    * The `args` if `system` is None.

    The unit is taken from:

    * The `scale` if not None (0=Bytes, 1=KiB, 2=MiB, ...).
    * The right scale from units previously retrieved from `args`.

    **Example usage**

    >>> naturalfilesize(-10)
    '-10 B'
    >>> naturalfilesize(0.233)
    '0.233 B'
    >>> naturalfilesize(1)
    '1 B'
    >>> naturalfilesize(69.5, format='{value:.2g} {unit}')
    '70 B'
    >>> naturalfilesize(999.9, format='{value:.0f}{unit}')
    '1000B'
    >>> naturalfilesize(1060)
    '1.04 kB'
    >>> naturalfilesize(1060, system='si')
    '1.06 KiB'
    >>> naturalfilesize(3210837)
    '3.06 MB'
    >>> naturalfilesize(3210837, scale=1, format='{value:.2f} {unit}')
    '3135.58 kB'
    >>> naturalfilesize(16262710, system=None, args={'base': 1000, 'units': ['B', 'K']})
    '1.63e+04 K'
    >>> naturalfilesize(314159265358979323846, system='gnu')
    '314 E'
    """
    return _naturalnumber(bytes, format=format, scale=scale, **(args[system] if system else args))


def naturalfrequency(
    hertz,
    format='{sign}{value:.3g} {unit}',
    scale=None,
    units=DEFAULT_FREQUENCY_UNITS
):  # pylint:disable=dangerous-default-value,redefined-builtin
    """
    Return a human readable representation of a frequency taking `hertz` as the frequency in Hz.

    The unit is taken from:

    * The `scale` if not None (0=bit/s, 1=kb/s, 2=Mb/s, ...).
    * The right scale from `units`.

    **Example usage**

    >>> naturalfrequency(-10)
    '-10 Hz'
    >>> naturalfrequency(0.233)
    '0.233 Hz'
    >>> naturalfrequency(69.5, format='{value:.2g} {unit}')
    '70 Hz'
    >>> naturalfrequency(999.9, format='{value:.0f}{unit}')
    '1000Hz'
    >>> naturalfrequency(1060)
    '1.06 kHz'
    >>> naturalfrequency(3210837)
    '3.21 MHz'
    >>> naturalfrequency(16262710, units=['Hertz', 'kilo Hertz'])
    '1.63e+04 kilo Hertz'
    >>> naturalfrequency(3210837, scale=1, format='{value:.2f} {unit}')
    '3210.84 kHz'
    """
    return _naturalnumber(hertz, base=1000, format=format, scale=scale, units=units)


def natural_int_key(text):
    """
    Function to be called as the key argument for list.sort() or sorted() in order to sort
    collections containing textual numbers on a more intuitive way.

    **Example usage**

    >>> sorted(['a26', 'a1', 'a4', 'a19', 'b2', 'a10', 'a3', 'b12'])
    ['a1', 'a10', 'a19', 'a26', 'a3', 'a4', 'b12', 'b2']
    >>> sorted(['a26', 'a1', 'a4', 'a19', 'b2', 'a10', 'a3', 'b12'], key=natural_int_key)
    ['a1', 'a3', 'a4', 'a10', 'a19', 'a26', 'b2', 'b12']
    """
    return [int(c) if c.isdigit() else c for c in DIGIT_REGEX.split(text)]


__all__ = _all.diff(globals())
